package be.iminds.ilabt.jfed.restapi.resource;

import be.iminds.ilabt.jfed.lowlevel.*;
import be.iminds.ilabt.jfed.restapi.JFedRestApiConfiguration;
import be.iminds.ilabt.jfed.restapi.dao.SpeaksForCredentialDoa;
import be.iminds.ilabt.jfed.restapi.json.SpeaksForCredential;
import be.iminds.ilabt.jfed.restapi.json.SpeaksForCredentialBuilder;
import be.iminds.ilabt.jfed.util.GeniUrn;
import be.iminds.ilabt.jfed.util.KeyUtil;
import com.codahale.metrics.annotation.Timed;
import io.swagger.annotations.Api;
import io.swagger.annotations.ApiOperation;
import io.swagger.annotations.ApiResponse;
import io.swagger.annotations.ApiResponses;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.servlet.http.HttpServletRequest;
import javax.validation.constraints.NotNull;
import javax.ws.rs.*;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.UriInfo;
import java.net.URI;
import java.net.URISyntaxException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;
import java.util.Date;
import java.util.List;

@Path("/speaksfor")
@Api(value="speaksfor")
@Produces(MediaType.APPLICATION_JSON)
public class SpeaksForCredentialResource {
    private static final Logger LOG = LoggerFactory.getLogger(SpeaksForCredentialResource.class);

    @Context
    UriInfo uriInfo;

    private final JFedRestApiConfiguration configuration;
    private final SpeaksForCredentialDoa speaksForCredentialDoa;

    public SpeaksForCredentialResource(JFedRestApiConfiguration configuration, SpeaksForCredentialDoa speaksForCredentialDoa) {
        this.configuration = configuration;
        this.speaksForCredentialDoa = speaksForCredentialDoa;
    }

    //request speaksfor for yourself
    @GET
    @Path("")
    @Produces(MediaType.APPLICATION_JSON)
    @ApiOperation("Request a SpeaksFor certificate for yourself")
    @ApiResponses(value= {
            @ApiResponse(code=403, message = "You do not have the necessary permissions to perform this action")
    })
    @Timed
    public SpeaksForCredential get(@Context HttpServletRequest request) {
        GeniUrn authenticatedUser = configuration.getAuthenticatedUser(request);
        if (authenticatedUser == null) {
            throw new WebApplicationException("The SSL connection did not correctly authenticate you as a trusted user. Did you use SSL client authentication?",
                    Response.Status.FORBIDDEN);
        }

        return get(authenticatedUser.getEncodedTopLevelAuthority(), authenticatedUser.getEncodedResourceName(), request);
    }

    @GET
    @Path("{tld}/{username}")
    @Produces(MediaType.APPLICATION_JSON)
    @ApiOperation("Request a SpeaksFor certificate for somebody else")
    @ApiResponses(value= {
            @ApiResponse(code=403, message = "You do not have the necessary permissions to perform this action")
    })
    @Timed
    public SpeaksForCredential get(@NotNull @PathParam("tld") String tld,
                                   @NotNull @PathParam("username") String username,
                                   @Context HttpServletRequest request) {
        GeniUrn requestedUserUrn = GeniUrn.createGeniUrnFromEncodedParts(tld, "user", username);

        GeniUrn authenticatedUser = configuration.getAuthenticatedUser(request);
        if (authenticatedUser == null) {
            throw new WebApplicationException("The SSL connection did not correctly authenticate you as a trusted user. Did you use SSL client authentication?",
                    Response.Status.FORBIDDEN);
        }

        authenticatedUser = configuration.modifyAuthenticatedUser(request, authenticatedUser, requestedUserUrn);
        if (!authenticatedUser.equals(requestedUserUrn) && !configuration.hasAccess(JFedRestApiConfiguration.Access.ADMIN, request)) {
            LOG.warn("FORBIDDING ACCESS: You are authenticated as user "+authenticatedUser+". You cannot request the speaksFor credential for "+requestedUserUrn);
            throw new WebApplicationException("You are authenticated as user "+authenticatedUser+". You cannot request the speaksFor credential for "+requestedUserUrn,
                    Response.Status.FORBIDDEN);
        }

        GeniUrn serviceUser = configuration.getUser().getUserUrn();
        SpeaksForCredential res = speaksForCredentialDoa.find(serviceUser, requestedUserUrn);

        if (res == null) {
            throw new WebApplicationException("No valid speaksFor credential known for "+requestedUserUrn,
                    Response.Status.NOT_FOUND);
        }
        return res;
    }

    @POST
    @Consumes(MediaType.APPLICATION_XML)
    @ApiOperation("Register a SpeaksFor credential to the server")
    @ApiResponses(value= {
            @ApiResponse(code=403, message = "You are not correctly authenticated, or the provided SpeaksFor credential is for another user.")
    })
    @Timed
    public Response postSpeaksForCredential(String speaksForCredentialXml, @Context HttpServletRequest request) {
        GeniUrn authenticatedUser = configuration.getAuthenticatedUser(request);
        if (authenticatedUser == null) {
            throw new WebApplicationException("The SSL connection did not correctly authenticate you as a trusted user. Did you use SSL client authentication?",
                    Response.Status.FORBIDDEN);
        }

        AnyCredential anyCredential;
        try {
            anyCredential = AnyCredential.createAutoDetect("credential received from "+authenticatedUser, speaksForCredentialXml);
        } catch (CredentialException e) {
            LOG.error("Error parsing credential", e);
            throw new WebApplicationException("Error parsing credential", Response.Status.BAD_REQUEST);
        }

        GeniUrn spokenForUser = SpeaksForCredentialResource.checkSpeaksForCredential(anyCredential, request, configuration);

        authenticatedUser = configuration.modifyAuthenticatedUser(request, authenticatedUser, spokenForUser);

        if (!authenticatedUser.equals(spokenForUser) && !configuration.hasAccess(JFedRestApiConfiguration.Access.ADMIN, authenticatedUser)) {
            LOG.warn("FORBIDDING ACCESS: You are authenticated as user "+authenticatedUser+". You cannot post a speaksFor credential for "+spokenForUser);
            throw new WebApplicationException("You are authenticated as user "+authenticatedUser+". You cannot post a speaksFor credential for "+spokenForUser,
                    Response.Status.FORBIDDEN);
        }

        GeniUser serviceUser = configuration.getUser();

        speaksForCredentialDoa.upsert(serviceUser.getUserUrn(), spokenForUser, anyCredential.getExpiresDate(), speaksForCredentialXml);

        URI uri = getUri(spokenForUser);

        return Response.created(uri).build();
    }


    /** returns spokenForUser
     * throws errors when something is wrong
     * */
    public static GeniUrn checkSpeaksForCredential(AnyCredential anyS4Credential, HttpServletRequest request, JFedRestApiConfiguration configuration) {
        if (anyS4Credential == null)
            throw new  WebApplicationException("Experiment.speaksForCredential is not a valid credential. (got null after parsing)", Response.Status.BAD_REQUEST);

        if (!(anyS4Credential instanceof AbacCredential)) {
            if (anyS4Credential instanceof SfaCredential && !((SfaCredential)anyS4Credential).isSpeaksFor()) {
                throw new WebApplicationException("Experiment.speaksForCredential is a valid " + anyS4Credential.getClass().getName() + " credential, but only ABAC speaksFor credentials are currently supported. " +
                        "Additionally, the SfaCredential does not appear to be a speaksFor credential.",
                        Response.Status.BAD_REQUEST);
            }

            throw new  WebApplicationException("Experiment.speaksForCredential is a valid "+anyS4Credential.getClass().getName()+" credential, but only AbacCredential is currently supported.",
                    Response.Status.BAD_REQUEST);
        }

        AbacCredential s4abac = (AbacCredential) anyS4Credential;

        if (!s4abac.isSpeaksFor()) {
            throw new WebApplicationException("Experiment.speaksForCredential is a valid ABAC credential, but it is not a speaksFor credential (which is required).",
                    Response.Status.BAD_REQUEST);
        }

        Date now = new Date();
        if (s4abac.getExpiresDate() != null && s4abac.getExpiresDate().before(now)) {
            throw new WebApplicationException("Experiment.speaksForCredential expired at "+s4abac.getExpires(),
                    Response.Status.BAD_REQUEST);
        }

        GeniUser serviceUser = configuration.getUser();
        X509Certificate serviceUserCert = serviceUser.getClientCertificateChain().get(0);

        String serviceUserId = AbacCredential.generateKeyId(serviceUserCert);
        String speakingId = s4abac.getTailKeyId();

        if (!speakingId.equals(serviceUserId)) {
            String serviceUserCertPem = KeyUtil.x509certificateToPem(serviceUserCert);
            throw new WebApplicationException("Experiment.speaksForCredential is a valid ABAC speaksFor credential, however, it does not allow speaking by the correct user. You need to allow the following user to speak for you to use this service:\n"+serviceUserCertPem,
                    Response.Status.BAD_REQUEST);
        }

        GeniUrn authenticatedUser = configuration.getAuthenticatedUser(request);
        if (authenticatedUser == null) {
            throw new WebApplicationException("The SSL connection did not correctly authenticate you as a trusted user. Did you use SSL client authentication?",
                    Response.Status.FORBIDDEN);
        }

//        PublicKey spokenForPubKey = s4abac.getSpokenForPubKey();
        GeniUrn spokenForUser = s4abac.getSpokenForUrn();
        if (spokenForUser == null) {
            throw new WebApplicationException("Experiment.speaksForCredential is a valid ABAC speaksFor credential, " +
                    "however the speaks_for user could not be derived from it.",
                    Response.Status.BAD_REQUEST);
        }

        authenticatedUser = configuration.modifyAuthenticatedUser(request, authenticatedUser, spokenForUser);

        if (!authenticatedUser.equals(spokenForUser) && !configuration.hasAccess(JFedRestApiConfiguration.Access.ADMIN, request)) {
            //HACK: config allows mapping users. Useful for testing.
            String mappedUser = configuration.getTestuserMapping() == null ? null : configuration.getTestuserMapping().get(authenticatedUser.toString());
            LOG.debug("User "+authenticatedUser+" mapped to "+mappedUser);

            if (mappedUser == null || !mappedUser.equals(spokenForUser.toString())) {
                throw new WebApplicationException("Access denied: you may not use another users speaksFor credential. " +
                        "Details: Experiment.speaksForCredential is a valid ABAC speaksFor credential, " +
                        "however it authorizes this service to speak for user \"" + spokenForUser + "\", " +
                        "while you are authenticated as user \"" + authenticatedUser + "\". ",
                        Response.Status.FORBIDDEN);
            }
        }

        //check if all certificates are valid at this time
        List<X509Certificate> signerCerts;
        try {
            signerCerts = s4abac.getSignerCertificates();
        } catch (CredentialException e) {
            throw new WebApplicationException("Invalid speaks_for credential", e, Response.Status.BAD_REQUEST);
        }

        for (X509Certificate certificate : signerCerts) {
            checkValidSpeaksForSignerX509Certificate(certificate);

            //TODO: use something like the code below
            //not that bad if this is not implemented: the AM's will check this.
            //advantage of implementing it here: better error message for users
//                if (!trustStore.isTrustedAuthority(certificate)) {
//                    throw new WebApplicationException("The authority signing the certificate is not trusted. tld="+singerIssuerTld, Response.Status.BAD_REQUEST);
//                }


            //TODO and check if it is trusted for what it says   (ex: wall2 is not trusted to say anything about geni users)
        }

        return spokenForUser;
    }

    private static void checkValidSpeaksForSignerX509Certificate(X509Certificate certificate) {
        try {
            certificate.checkValidity();
        } catch (CertificateExpiredException e) {
            LOG.error("The signer certificate in the speaks_for credential has expired."+
                    "(NotAfter= "+certificate.getNotAfter()+")", e);
            throw new WebApplicationException("The signer certificate in the speaks_for credential has expired. " +
                    "(NotAfter= "+certificate.getNotAfter()+")", e, Response.Status.FORBIDDEN);
        } catch (CertificateNotYetValidException e) {
            throw new WebApplicationException("The signer certificate in the speaks_for credential is not yet valid " +
                    "(NotBefore="+certificate.getNotBefore()+")", e, Response.Status.FORBIDDEN);
        }
    }


    public URI getUri(GeniUrn speakingFor) {
        String base = configuration == null || configuration.getBaseUrl() == null || configuration.getBaseUrl().trim().isEmpty() ?
                null : configuration.getBaseUrl();
        if (base != null && !base.endsWith("/")) base += "/";

        assert speakingFor != null;

        URI uri = null;
        try {
            uri = new URI(base + "speaksfor/"+speakingFor.getEncodedTopLevelAuthority_withoutSubAuth()+"/"+speakingFor.getEncodedResourceName());
        } catch (URISyntaxException e) {
            LOG.error("Error when adding URI", e);
            //throw new WebApplicationException("Error when adding URI", Response.Status.INTERNAL_SERVER_ERROR);
            //ignore
        }

        return uri;
    }

    public SpeaksForCredentialBuilder addUri(SpeaksForCredentialBuilder speaksForCredentialBuilder) {
        GeniUrn speakingFor = speaksForCredentialBuilder.getSpeakingFor();
        assert speakingFor != null;

        speaksForCredentialBuilder.setUri(getUri(speakingFor));

        return speaksForCredentialBuilder;
    }
}
